# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from copy import copy
from ..spec import spec
from ..decorators import check


# Module API

@check('missing-header')
def missing_header(errors, columns, sample=None):
    for column in copy(columns):
        if 'header' not in column:
            # Add error
            message = spec['errors']['missing-header']['message']
            message = message.format(column_number=column['number'])
            errors.append({
                'code': 'missing-header',
                'message': message,
                'row-number': None,
                'column-number': column['number'],
            })
            # Remove column
            columns.remove(column)



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from ..spec import spec
from ..decorators import check


# Module API

@check('unique-constraint')
def unique_constraint(errors, columns, row_number, state):
    rindexes = state.setdefault('cache', {})
    for column in columns:
        if len(column) == 4:
            if column['field'].constraints.get('unique'):
                rindex = rindexes.setdefault(column['number'], {})
                references = rindex.setdefault(column['value'], [])
                if references:
                    # Add error
                    message = spec['errors']['unique-constraint']['message']
                    message = message.format(
                        row_numbers=', '.join(map(str, references + [row_number])),
                        column_number=column['number'])
                    errors.append({
                        'code': 'unique-constraint',
                        'message': message,
                        'row-number': row_number,
                        'column-number': column['number'],
                    })
                references.append(row_number)



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import six
import jsontableschema
from tabulator import Stream
from jsontableschema import Schema, validate
from ..decorators import preset


# Module API

@preset('table')
def table(source, schema=None, **options):
    warnings = []
    tables = []

    # Ensure not a datapackage
    if isinstance(source, six.string_types):
        if source.endswith('datapackage.json'):
            warnings.append('Use "datapackage" preset for Data Packages')

    # Prepare schema
    if not warnings:
        if schema is not None:
            # TODO: after tableschema-v1 will be ready
            # - we should use Schema(strict=False) to handle schema errors on inspection
            # - it means we don't need to validate schema here, mess with helpers etc
            try:
                from jsontableschema.helpers import load_json_source
                loaded_descriptor = load_json_source(schema)
                validate(loaded_descriptor, no_fail_fast=True)
                schema = Schema(loaded_descriptor)
            except jsontableschema.exceptions.MultipleInvalid as exception:
                for error in exception.errors:
                    warnings.append(
                        'Table schema "%s" has a validation error "%s"' %
                        (schema, str(error).splitlines()[0]))
            except Exception as error:
                warnings.append(
                    'Table Schema "%s" has a loading error "%s"' %
                    (schema, error))

    # Add table
    if not warnings:
        options.setdefault('headers', 1)
        tables.append({
            'source': str(source),
            'stream': Stream(source, **options),
            'schema': schema,
            'extra': {},
        })

    return warnings, tables



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import json
from ..spec import spec
from ..decorators import check


# Module API

@check('duplicate-row')
def duplicate_row(errors, columns, row_number, state):
    rindex = state.setdefault('rindex', {})
    try:
        pointer = hash(json.dumps(list(column.get('value') for column in columns)))
        references = rindex.setdefault(pointer, [])
    except TypeError:
        pointer = None
    if pointer:
        if references:
            # Add error
            message = spec['errors']['duplicate-row']['message']
            message = message.format(
                row_number=row_number,
                row_numbers=', '.join(map(str, references)))
            errors.append({
                'code': 'duplicate-row',
                'message': message,
                'row-number': row_number,
                'column-number': None,
            })
            # Clear columns
            del columns[:]
        references.append(row_number)



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals


# Module API

def preset(name):
    """https://github.com/frictionlessdata/goodtables-py#custom-presets
    """
    def decorator(func):
        func.preset = {
            'name': name,
        }
        return func
    return decorator


def check(code, type=None, context=None, before=None, after=None):
    """https://github.com/frictionlessdata/goodtables-py#custom-checks
    """
    def decorator(func):
        func.check = {
            'code': code,
            'type': type,
            'context': context,
            'before': before,
            'after': after,
        }
        return func
    return decorator



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import re
from copy import copy
from ..spec import spec
from ..decorators import check


# Module API

@check('non-matching-header')
def non_matching_header(errors, columns, sample=None, order_fields=False):

    # No field ordering
    if not order_fields:
        for column in copy(columns):
            if len(column) == 3:
                if column['header'] != column['field'].name:
                    # Add error
                    message = spec['errors']['non-matching-header']['message']
                    message = message.format(
                        column_number=column['number'],
                        field_name=column['field'].name)
                    errors.append({
                        'code': 'non-matching-header',
                        'message': message,
                        'row-number': None,
                        'column-number': column['number'],
                    })
                    if _slugify(column['header']) != _slugify(column['field'].name):
                        # Remove column
                        columns.remove(column)

    # Field ordering
    else:
        # Update fields order to maximally match headers order
        headers = [column.get('header') for column in columns]
        for index, header in enumerate(headers):
            if header is None:
                continue
            field_name = _get_field_name(columns[index])
            # Column header and field_name are different
            if _slugify(header) != _slugify(field_name):
                # If there is the match in next columns - swap fields
                for column in columns[index+1:]:
                    # We've found field matching given header
                    if _slugify(header) == _slugify(_get_field_name(column)):
                        _swap_fields(columns[index], column)
                        break
                    if field_name:
                        # Given field matches some header also swap
                        if _slugify(field_name) == _slugify(column.get('header')):
                            _swap_fields(columns[index], column)
        # Run check with no field ordering
        non_matching_header(errors, columns)


# Internal


def _get_field_name(column):
    field_name = None
    if 'field' in column:
        field_name = column['field'].name
    return field_name


def _slugify(string):
    if string is None:
        return None
    string = re.sub(r'[\W_]+', '', string)
    string = string.lower()
    return string


def _swap_fields(column1, column2):
    field1 = column1.get('field')
    column1['field'] = column2.get('field')
    column2['field'] = field1
    if column1['field'] is None:
        del column1['field']
    if column2['field'] is None:
        del column2['field']



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from ..spec import spec
from ..decorators import check


# Module API

@check('blank-header')
def blank_header(errors, columns, sample=None):
    for column in columns:
        if 'header' in column:
            if not column['header']:
                # Add error
                message = spec['errors']['blank-header']['message']
                message = message.format(column_number=column['number'])
                errors.append({
                    'code': 'blank-header',
                    'message': message,
                    'row-number': None,
                    'column-number': column['number'],
                })



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals


# Module API

from .blank_header import blank_header
from .blank_row import blank_row
from .duplicate_row import duplicate_row
from .duplicate_header import duplicate_header
from .enumerable_constraint import enumerable_constraint
from .extra_header import extra_header
from .extra_value import extra_value
from .maximum_constraint import maximum_constraint
from .maximum_length_constraint import maximum_length_constraint
from .minimum_constraint import minimum_constraint
from .minimum_length_constraint import minimum_length_constraint
from .missing_header import missing_header
from .missing_value import missing_value
from .non_castable_value import non_castable_value
from .non_matching_header import non_matching_header
from .pattern_constraint import pattern_constraint
from .required_constraint import required_constraint
from .unique_constraint import unique_constraint



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from ..spec import spec
from ..decorators import check


# Module API

@check('maximum-length-constraint')
def maximum_length_constraint(errors, columns, row_number, state=None):
    for column in columns:
        if len(column) == 4:
            valid = column['field'].test_value(column['value'], constraint='maxLength')
            if not valid:
                # Add error
                message = spec['errors']['maximum-length-constraint']['message']
                message = message.format(
                    value=column['value'],
                    row_number=row_number,
                    column_number=column['number'],
                    constraint=column['field'].constraints['maxLength'])
                errors.append({
                    'code': 'maximum-length-constraint',
                    'message': message,
                    'row-number': row_number,
                    'column-number': column['number'],
                })



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from tabulator import Stream
from jsontableschema import Schema
from jsontableschema.exceptions import SchemaValidationError
from datapackage import DataPackage
from ..decorators import preset


# Module API

@preset('datapackage')
def datapackage(source, **options):
    warnings = []
    tables = []

    # Load datapackage
    try:
        datapackage = DataPackage(source, **options)
    except Exception as error:
        warnings.append(
            'Data Package "%s" has a loading error "%s"' %
            (source, error))

    # Validate datapackage
    if not warnings:
        for error in datapackage.iter_errors():
            warnings.append(
                'Data Package "%s" has a validation error "%s"' %
                (source, str(error).splitlines()[0]))

    # Extract datapackage tables
    if not warnings:
        for resource in datapackage.resources:
            # TODO: after datapackage-v1 will be ready
            # - we should use `resource.tabular` to filter tabular resources
            # - we don't need to validate schema here because of dereferencing
            path = resource.remote_data_path or resource.local_data_path
            try:
                schema = Schema(resource.descriptor['schema'])
            except SchemaValidationError as error:
                warnings.append(
                    'Data Package "%s" has a validation error "%s"' %
                    (source, str(error).splitlines()[0]))
                continue
            except Exception:
                continue
            tables.append({
                'source': path,
                'stream': Stream(path, headers=1),
                'schema': schema,
                'extra': {
                    'datapackage': str(source),
                },
            })

    return warnings, tables



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import six
import inspect
import datetime
import operator
import tabulator
from copy import copy
from itertools import chain
from functools import partial
from six.moves import zip_longest
from jsontableschema import Schema, infer
from multiprocessing.pool import ThreadPool
from . import presets as presets_module
from . import checks as checks_module
from . import exceptions
from . import config
from .spec import spec


# Module API

class Inspector(object):

    # Public

    def __init__(self,
                 checks='all',
                 error_limit=1000,
                 table_limit=10,
                 row_limit=1000,
                 infer_schema=False,
                 infer_fields=False,
                 order_fields=False,
                 custom_presets=[],
                 custom_checks=[]):
        """https://github.com/frictionlessdata/goodtables-py#inspector
        """

        # Set attributes
        self.__error_limit = error_limit
        self.__table_limit = table_limit
        self.__row_limit = row_limit
        self.__infer_schema = infer_schema
        self.__presets = _prepare_presets(copy(custom_presets))
        self.__checks = _prepare_checks(checks, copy(custom_checks),
            order_fields=order_fields, infer_fields=infer_fields)

    def inspect(self, source, preset='table', **options):
        """https://github.com/frictionlessdata/goodtables-py#inspector
        """

        # Start timer
        start = datetime.datetime.now()

        # Prepare preset
        try:
            preset_func = self.__presets[preset]
            if preset == 'nested':
                options['presets'] = self.__presets
        except KeyError:
            message = 'Preset "%s" is not registered' % preset
            raise exceptions.GoodtablesException(message)

        # Prepare tables
        warnings, tables = preset_func(source, **options)
        if len(tables) > self.__table_limit:
            warnings.append(
                'Dataset inspection has reached %s table(s) limit' %
                (self.__table_limit))
            tables = tables[:self.__table_limit]

        # Collect table reports
        table_reports = []
        if tables:
            tasks = []
            pool = ThreadPool(processes=len(tables))
            for table in tables:
                tasks.append(pool.apply_async(self.__inspect_table, (table,)))
            for task in tasks:
                table_warnings, table_report = task.get()
                warnings.extend(table_warnings)
                table_reports.append(table_report)

        # Stop timer
        stop = datetime.datetime.now()

        # Compose report
        report = {
            'time': round((stop - start).total_seconds(), 3),
            'valid': all(item['valid'] for item in table_reports),
            'error-count': sum(len(item['errors']) for item in table_reports),
            'table-count': len(tables),
            'tables': table_reports,
            'warnings': warnings,
        }

        return report

    # Internal

    def __inspect_table(self, table):

        # Start timer
        start = datetime.datetime.now()

        # Prepare vars
        errors = []
        warnings = []
        headers = []
        row_number = 0
        fatal_error = False
        checks = copy(self.__checks)
        source = table['source']
        stream = table['stream']
        schema = table['schema']
        extra = table['extra']

        # Prepare table
        try:
            stream.open()
            sample = stream.sample
            headers = stream.headers
            if headers is None:
                headers = [None] * len(sample[0]) if sample else []
            if _filter_checks(checks, type='schema'):
                if schema is None and self.__infer_schema:
                    schema = Schema(infer(headers, sample))
            if schema is None:
                checks = _filter_checks(checks, type='schema', inverse=True)
        except Exception as exception:
            fatal_error = True
            error = _compose_error_from_exception(exception)
            errors.append(error)

        # Prepare columns
        if not fatal_error:
            columns = []
            fields = [None] * len(headers)
            if schema is not None:
                fields = schema.fields
            iterator = zip_longest(headers, fields, fillvalue=_FILLVALUE)
            for number, (header, field) in enumerate(iterator, start=1):
                column = {'number': number}
                if header is not _FILLVALUE:
                    column['header'] = header
                if field is not _FILLVALUE:
                    column['field'] = field
                columns.append(column)

        # Head checks
        if not fatal_error:
            if None not in headers:
                head_checks = _filter_checks(checks, context='head')
                for check in head_checks:
                    if not columns:
                        break
                    check['func'](errors, columns, sample)
                for error in errors:
                    error['row'] = None

        # Body checks
        if not fatal_error:
            states = {}
            colmap = {column['number']: column for column in columns}
            body_checks = _filter_checks(checks, context='body')
            with stream:
                extended_rows = stream.iter(extended=True)
                while True:
                    try:
                        row_number, _, row = next(extended_rows)
                    except StopIteration:
                        break
                    except Exception as exception:
                        fatal_error = True
                        error = _compose_error_from_exception(exception)
                        errors.append(error)
                        break
                    columns = []
                    iterator = zip_longest(headers, row, fillvalue=_FILLVALUE)
                    for number, (header, value) in enumerate(iterator, start=1):
                        colref = colmap.get(number, {})
                        column = {'number': number}
                        if header is not _FILLVALUE:
                            column['header'] = colref.get('header', header)
                        if 'field' in colref:
                            column['field'] = colref['field']
                        if value is not _FILLVALUE:
                            column['value'] = value
                        columns.append(column)
                    for check in body_checks:
                        if not columns:
                            break
                        state = states.setdefault(check['code'], {})
                        check['func'](errors, columns, row_number, state)
                    for error in reversed(errors):
                        if 'row' in error:
                            break
                        error['row'] = row
                    if row_number >= self.__row_limit:
                        warnings.append(
                            'Table "%s" inspection has reached %s row(s) limit' %
                            (source, self.__row_limit))
                        break
                    if len(errors) >= self.__error_limit:
                        warnings.append(
                            'Table "%s" inspection has reached %s error(s) limit' %
                            (source, self.__error_limit))
                        break

        # Stop timer
        stop = datetime.datetime.now()

        # Compose report
        headers = headers if None not in headers else None
        errors = errors[:self.__error_limit]
        errors = _sort_errors(errors)
        report = copy(extra)
        report.update({
            'time': round((stop - start).total_seconds(), 3),
            'valid': not bool(errors),
            'error-count': len(errors),
            'row-count': row_number,
            'headers': headers,
            'source': source,
            'errors': errors,
        })

        return warnings, report


# Internal

_FILLVALUE = '_FILLVALUE'


def _prepare_presets(custom):

    # Prepare presets
    presets = {}
    for preset in chain(vars(presets_module).values(), custom):
        descriptor = getattr(preset, 'preset', None)
        if descriptor:
            presets[descriptor['name']] = preset

    return presets


def _prepare_checks(setup, custom, order_fields, infer_fields):

    # Prepare errors/checkmap
    errors = []
    checkmap = {}
    for code in config.CHECKS:
        error = copy(spec['errors'][code])
        error.update({'code': code})
        errors.append(error)
    for check in chain(vars(checks_module).values(), custom):
        desc = getattr(check, 'check', None)
        if desc:
            errormap = {error['code']: index for index, error in enumerate(errors)}
            if desc['before'] in errormap:
                errors.insert(errormap[desc['before']], desc)
            if desc['after'] in errormap:
                errors.insert(errormap[desc['after']] + 1, desc)
            checkmap[desc['code']] = check

    # Prepare checks
    checks = []
    for error in errors:
        if error['code'] in checkmap:
            checks.append({
                'func': checkmap[error['code']],
                'code': error['code'],
                'type': error['type'],
                'context': error['context'],
            })

    # Filter structure checks
    if setup == 'structure':
        checks = _filter_checks(checks, type='structure')

    # Filter schema checks
    elif setup == 'schema':
        checks = _filter_checks(checks, type='schema')

    # Filter granular checks
    elif isinstance(setup, dict):
        default = True not in setup.values()
        checks = [check for check in checks
            if setup.get(check['code'], default)]

    # Unknown filter
    elif setup != 'all':
        message = 'Checks filter "%s" is not supported' % setup
        raise exceptions.GoodtablesException(message)

    # Bind options
    for check in checks:
        if six.PY2:
            parameters, _, _, _ = inspect.getargspec(check['func'])
        else:
            parameters = inspect.signature(check['func']).parameters
        if 'order_fields' in parameters:
            check['func'] = partial(check['func'], order_fields=order_fields)
        if 'infer_fields' in parameters:
            check['func'] = partial(check['func'], infer_fields=infer_fields)

    return checks


def _filter_checks(checks, type=None, context=None, inverse=False):

    # Filted checks
    result = []
    comparator = operator.ne
    if inverse:
        comparator = operator.eq
    for check in checks:
        if type and comparator(check['type'], type):
            continue
        if context and comparator(check['context'], context):
            continue
        result.append(check)

    return result


def _compose_error_from_exception(exception):
    code = 'source-error'
    message = str(exception)
    if isinstance(exception, tabulator.exceptions.SourceError):
        code = 'source-error'
    elif isinstance(exception, tabulator.exceptions.SchemeError):
        code = 'scheme-error'
    elif isinstance(exception, tabulator.exceptions.FormatError):
        code = 'format-error'
    elif isinstance(exception, tabulator.exceptions.EncodingError):
        code = 'encoding-error'
    elif isinstance(exception, tabulator.exceptions.IOError):
        code = 'io-error'
    elif isinstance(exception, tabulator.exceptions.HTTPError):
        code = 'http-error'
    return {
        'row': None,
        'code': code,
        'message': message,
        'row-number': None,
        'column-number': None,
    }


def _sort_errors(errors):
    return sorted(errors, key=lambda error:
        (error['row-number'] or 0, error['column-number']))



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from copy import copy
from jsontableschema import Schema, infer
from ..spec import spec
from ..decorators import check


# Module API

@check('extra-header')
def extra_header(errors, columns, sample, infer_fields=False):
    for column in copy(columns):
        if 'field' not in column:
            # Infer field
            if infer_fields:
                column_sample = []
                for row in sample:
                    value = None
                    if len(row) > column['number']:
                        value = row[column['number']]
                    column_sample.append(value)
                descriptor = infer([column['header']], column_sample)
                column['field'] = Schema(descriptor).fields[0]
            # Add error/remove column
            else:
                message = spec['errors']['extra-header']['message']
                message = message.format(column_number=column['number'])
                errors.append({
                    'code': 'extra-header',
                    'message': message,
                    'row-number': None,
                    'column-number': column['number'],
                })
                columns.remove(column)
